/* PopAccount.cc - source file for the mailfilter program
 * Copyright (c) 2000  Andreas Bauer <baueran@in.tum.de>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
 * USA.
 */

#include <strstream>
#include <fstream>
#include <string>
#include <vector>
#include <sys/time.h>
#include <stdlib.h>
#include <netinet/in.h>
#include <unistd.h>
#include <arpa/inet.h>
#include <netdb.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/stat.h>
#include <stdio.h>
#include <ctype.h>
#include <regex.h>
#include <fcntl.h>
#include <string.h>
#include "Account.hh"
#include "PopAccount.hh"
#include "Preferences.hh"
#include "Feedback.hh"
#include "mailfilter.hh"
#include "config.h"


using namespace std;
using namespace acc;

namespace pop {

  PopAccount::PopAccount(string server, string user, string password, int port, pref::Preferences* newPrefs) {
    this->server = server;
    this->user = user;
    this->pass = password;
    this->port = port;
    prefs = newPrefs;
    mySocket = 0;
    report = new fb::Feedback(prefs->getLogfile(), prefs->getMode());
  }
  
  
  PopAccount::~PopAccount() {
    // No need to delete prefs, as we only point to the prefs object created in mailfilter.cc
    // report, on the other hand, was assigned memory to so we better clean up.
    delete(report);      
  }
  
  
  int PopAccount::check() {
    char cmd[32];
    string result;
    int messages = 0, size = 0, connectError = 0, loginError = 0;
    
    // Establish server connection and login
    if ( ((connectError = connectSocket()) == 0) && ((loginError = login()) == 0) ) {
      
      // Store message headers
      try {
	if (sendCommand("STAT\r\n") <= 0) {
	  logout();
	  disconnectSocket();
	  return(CMD_STAT_FAILED);
	}
	else {
	  result = receiveResult();
	  
	  if (serverResult(result) == FALSE) {
	    logout();
	    disconnectSocket();
	    return(CMD_STAT_FAILED);
	  }
	  else {
	    messages = atoi(getWord(result, 1).c_str());
	    char status[MAX_BYTES];
#ifdef HAVE_SNPRINTF
	    snprintf(status, sizeof(status), "%s: Examining %d message(s).\n", PACKAGE, messages);
#else
	    sprintf(status, "%s: Examining %d message(s).\n", PACKAGE, messages);
#endif
	    report->message(status);
	  
	    // Store the header for each message in the mail box
	    for (int i = 0; i < messages; i++) {

	      // Get message size
#ifdef HAVE_SNPRINTF
	      snprintf(cmd, sizeof(cmd), "LIST %d\r\n", i + 1);
#else
	      sprintf(cmd, "LIST %d\r\n", i + 1);
#endif
	      if (sendCommand(cmd) <= 0) {
		logout();
		disconnectSocket();
		return(CMD_LIST_FAILED);
	      }
	      else {
		result = receiveResult();

		if (serverResult(result) == FALSE) {
		  disconnectSocket();
		  return(CMD_LIST_FAILED);
		}
		else
		  size = atoi(getWord(result, 2).c_str());
	      }

	      // Get header
#ifdef HAVE_SNPRINTF
	      snprintf(cmd, sizeof(cmd), "TOP %d 0\r\n", i + 1);
#else
	      sprintf(cmd, "TOP %d 0\r\n", i + 1);
#endif
	      if (sendCommand(cmd) <= 0) {
		logout();
		disconnectSocket();
		return(CMD_TOP_FAILED);
	      }
	      else {
		// We now receive the e-mail header by assigning the server command to the function.
		// This is necessary cause IMail under Windows NT has the annoying behaviour of
		// sending the response to TOP in two seperate streams and somehow I have to inform
		// receiveResult that we are waiting for the second stream, even though it could
		// all be handled via one stream easily.
		result = receiveResult(cmd);
		
		// Store header content, the number of the message and its size in bytes
		if (serverResult(result) == TRUE) {
		  if (Account::storeMessageHeader(result, i + 1, size) < 0) {
		    logout();
		    disconnectSocket();
		    return(MALFORMED_HEADER_FAILURE);
		  }
		}
		else {
		  logout();
		  disconnectSocket();
		  return(CMD_TOP_FAILED);
		}
	      }
	    } // for
	  } // else
	} // else
      } // try
      catch(...) {
	disconnectSocket();
	throw;
      }
            
      // Now filter messages and delete spam and unwanted email
      vector<Header>::iterator curMessage = Account::headers.begin();
      while(curMessage != Account::headers.end()) {
	// Check size (and delete)
	if ( (prefs->getMaxsize() > 0) && (curMessage->size > prefs->getMaxsize()) ) {
	  if (deleteMessage(curMessage->number) < 0)
	    return(CMD_DELE_FAILED);
	  else
	    report->message((string)PACKAGE + (string)": Deleted " + curMessage->sender.c_str() + (string)": " + curMessage->subject.c_str() + (string)", " + curMessage->date.c_str() + (string)".\n");
	}
	// Check filter rules (and delete)
	else {
	  vector<Line>::iterator curLine = curMessage->lines.begin();
	  
	  while(curLine != curMessage->lines.end()) {
	    regex_t compFilter;
	    vector<string>::iterator curFilter = prefs->getFilters()->begin();

	    // regfree(&compFilter);      // Not so sure if on which platforms regfree breaks and which might need it.
	                                  // I know that Linux with glibc 2.1.1 doesn't particularly like it...

	    while (curFilter != prefs->getFilters()->end()) {
	      string currentLine = curLine->descr.c_str();
	      currentLine.append(": ");
	      currentLine.append(curLine->content.c_str());
	      
	      if (prefs->getIcase() == FALSE)
		regcomp(&compFilter, curFilter->c_str(), REG_ICASE);
	      else
		regcomp(&compFilter, curFilter->c_str(), 0);

	      // Delete spam mail here and write report
	      if (regexec(&compFilter, currentLine.c_str(), 0, NULL, 0) == 0) {
		try {
		  if (deleteMessage(curMessage->number) >= 0) {
		    report->message((string)PACKAGE + (string)": Deleted " + curMessage->sender.c_str() + (string)": " + curMessage->subject.c_str() + (string)", " + curMessage->date.c_str() + (string)".\n");
		    curFilter = prefs->getFilters()->end(); // Jump to the end of the filters! No more checking is needed.
		  }		
		  else {
		    logout();
		    disconnectSocket();
		    return(CMD_DELE_FAILED);
		  }
		}
		catch(...) {
		  disconnectSocket();
		  throw;
		}
	      }
	      // Mail not yet deleted, check if maybe next filter matches
	      else
		curFilter++;
	    }

	    curMessage->lines.erase(curLine);  // instead of curLine++
	  }
	}
	
	Account::headers.erase(curMessage);    // instead of curMessage++
      }

      // Logout and clean up connections
      try {
	logout();
	disconnectSocket();
	return 0;
      }
      catch(...) {
	disconnectSocket();
	throw;
      }
    }
    else if (loginError != 0) {
      disconnectSocket();
      return AUTHENTICATION_FAILURE;
    }
    else if (connectError != 0) {
      disconnectSocket();
      return connectError;
    }
    else {
      // Unknown error occured, throw exception
      disconnectSocket();
      throw UnknownError();
    }
  }
  
  
  // Creates a sockets and creates connection to the server
  // Returns 0 on success, or DNS_LOOKUP_FAILURE and SOCKET_CONNECTION_FAILURE on error
  int PopAccount::connectSocket() {
    struct sockaddr_in socketAddress;
    struct hostent *myHost;
    
    // Create socket and try to connect
    if ( (mySocket = socket(PF_INET, SOCK_STREAM, 0)) < 0)
      return(SOCKET_CONNECTION_FAILURE);
    else {
      myHost = gethostbyname(server.c_str());
      
      if (!myHost)
	return(DNS_LOOKUP_FAILURE);
      else {
	memset( (char*)&socketAddress, 0, sizeof(struct sockaddr_in) );
	socketAddress.sin_family = AF_INET;
	memcpy( &(socketAddress.sin_addr.s_addr), myHost->h_addr, myHost->h_length);
	socketAddress.sin_port=htons(port);
	
	if (connect(mySocket, (struct sockaddr*)&socketAddress, sizeof(struct sockaddr)) < 0)
	  return(SOCKET_CONNECTION_FAILURE);
      }
    }
    
    return 0;
  }
  
  
  // Login to pop server
  // Returns 0 on success, -1 otherwise
  int PopAccount::login() {
    char command[32];
    
    try {
      // Check whether we received the server's greeting message
      if (serverResult(receiveResult()) == TRUE) {
	
	// Send username
#ifdef HAVE_SNPRINTF
	snprintf(command, sizeof(command), "USER %s\r\n", user.c_str());
#else
	sprintf(command, "USER %s\r\n", user.c_str());
#endif
	if (sendCommand(command) > 0) {
	  if (serverResult(receiveResult()) == FALSE) {
	    disconnectSocket();
	    return -1;
	  }
	}
	else
	  return -1;
	
	// Send password
#ifdef HAVE_SNPRINTF
	snprintf(command, sizeof(command), "PASS %s\r\n", pass.c_str());
#else
	sprintf(command, "PASS %s\r\n", pass.c_str());
#endif
	if (sendCommand(command) > 0) { 
	  if (serverResult(receiveResult()) == FALSE) {
	    disconnectSocket();
	    return -1;
	  }
	}
	else
	  return -1;
	
	return 0;
      }
      else
	return -1;
    }
    catch(...) {
      throw;
    }
  }
  
  
  // Closes the socket
  int PopAccount::disconnectSocket() {
    return (close(mySocket));
  }


  // Disconnects from the server
  // Returns 0 on success, -1 otherwise
  int PopAccount::logout() {
    try {
      if ( (sendCommand("QUIT\r\n") > 0) && (serverResult(receiveResult()) == TRUE) )
	return 0;
      else
	return -1;
    }
    catch(...) {
      throw;
    }
  }
  
  
  // Sends a command to the socket (server)
  // Commands have to be < 32, otherwise -1 is returned
  // Returns bytes written on success, -1 otherwise
  int PopAccount::sendCommand(string command) {
    char cmd[32];
    int error = 0;

    if (command.length() <= 32) { 
      strcpy(cmd, command.c_str());

      if ( (error = write(mySocket, cmd, strlen(cmd))) == -1 )
	return -1;
      else
	return error;
    }
    else
      return -1;
  }


  // Receives the result from a previous command sent to the server.
  // Throws exception on communication errors.
  // Returns the received text as string object.
  //
  // Note: serverCommand is only needed for the TOP n m command and only because
  //       IMail on Windows NT sends two streams, one after another, to return
  //       the e-mail header of a message. Quite odd.
  string PopAccount::receiveResult(string serverCommand = "irrelevant") {
    static const int TIME_OUT = 10;
    char buffer[MAX_BYTES + 20];   // Leave some space for expansion
    int flags = 0, error = 0;
    struct timeval tv;
    fd_set rfds;
    string input;
    int counter = 0, bytes = 0, last = 0;
    
    // Memorize old flags of the open socket
    if ( (flags = fcntl(mySocket, F_GETFL, 0)) == -1 )
      throw IOException();

    // Change the socket communication stream to nonblocking
    if (fcntl(mySocket, F_SETFL, flags | O_NONBLOCK) == -1 )
      throw IOException();

    do {
      memset(buffer, 0, sizeof(buffer));
      last = bytes;
      tv.tv_sec = TIME_OUT;
      tv.tv_usec = 0;
      FD_ZERO(&rfds);
      FD_SET(mySocket, &rfds);
      
      // Only read if the socket is ready and sending data
      error = select(mySocket + 1, &rfds, NULL, NULL, &tv);

      if (error < 0)
	throw ConnectionTimeOut();
      else if (error > 0 && FD_ISSET(mySocket, &rfds)) {
	bytes = read(mySocket, buffer, MAX_BYTES);
	
	if (bytes > 0) {
	  buffer[bytes] = 0;
	  counter += bytes;

	  if (input.length() > 0)
	    input.append(buffer, bytes);
	  else
	    input = (string)buffer;
	}
	else if (bytes == 0)
	  break;
	else
	  throw IOException();
      }
    } while (   (error > 0)  &&
		(last > 0 && bytes > 0 && !isHeaderEnd(input))  ||
		( (last > 0 || bytes == MAX_BYTES) && !isHeaderEnd(input) )  ||
		( (serverCommand.find("TOP") == 0) && !isHeaderEnd(input) )     );   // This line only makes sense with the IMail server
                                                                                     // on Windows NT. See check() for more information.

    // Put line end behind the buffer
    input[counter] = 0;

    // Reset socket communication to blocking mode
    if ( (fcntl(mySocket, F_SETFL, flags)) == -1)
      throw IOException();
    
    // Return the server response
    return(input);
  }
  

  // Deletes a message on the POP3 server
  // Returns 0 on success, -1 otherwise
  int PopAccount::deleteMessage(int mess) {
    char cmd[32];
#ifdef HAVE_SNPRINTF
    snprintf(cmd, sizeof(cmd), "DELE %d\r\n", mess);
#else
    sprintf(cmd, "DELE %d\r\n", mess);
#endif
    try {
      if ( (sendCommand(cmd) > 0) && (serverResult(receiveResult()) == TRUE) )
	return 0;
      else
	return -1;
    }
    catch(...) {
      throw;
    }
  }

  
  // Returns the n-th word in a string,
  // e.g. getWord("My name is Harry", 1) would return "name".
  string PopAccount::getWord(const string word, int n) {
    int curSpace = 0, nextSpace = 0, wordEnd = 0;

    // Find beginning of word
    for (int i = 0; i < n; i++) {
      nextSpace = word.find(' ', curSpace);
      curSpace = nextSpace + 1;
    }
    
    // Either copy everything up to the next blank character, or up to the end of the line (-1 for the cr)
    if ( (wordEnd = word.find(' ', curSpace)) == -1 )
	wordEnd = word.length() - 1;

    return(word.substr(curSpace, wordEnd));
  }


  // Checks whether server message is an error message (-ERR) or a positive status report (+OK)
  bool PopAccount::serverResult(const string serverMessage) {
    if (serverMessage[0] == '+')
      return TRUE;
    else
      return FALSE;    
  }


  // Returns true if the string buffer contains the end of an e-mail message header
  // otherwise the function returns false.
  // Note: according to the POP3 specs the end of messages are seperated by <CR><LF><.><CR><LF>
  bool PopAccount::isHeaderEnd(const string buffer) {
    if (buffer.length() >= 4)
      for (unsigned int i = 0; i+4 <= buffer.length(); i++)
	if ( (i+4 <= buffer.length()) && (buffer[i] == '\r') && (buffer[i+1] == '\n') && (buffer[i+2] == '.') &&
	     (buffer[i+3] == '\r') && (buffer[i+4] == '\n') )
	  return TRUE;

    return FALSE;
  }

}
