/* mailfilter.cc - source file for the mailfilter program
 * Copyright (c) 2000  Andreas Bauer <baueran@in.tum.de>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
 * USA.
 */

#include <string>
#include <iostream>
#include <sys/time.h>
#include <unistd.h>
#include <sys/types.h>
#include <pwd.h>
#include <netdb.h>
#include <stdlib.h>
#include <unistd.h>

#ifdef HAVE_GETOPT_H
#include <getopt.h>
#else
#include <stdlib.h>
#include "getopt.h"
#endif

#include "config.h"
#include "mailfilter.hh"
#include "Preferences.hh"
#include "Feedback.hh"
#include "PopAccount.hh"

using namespace std;

int main(int argc, char *argv[]) {
  int error = 0, c = 0, option_index = 0, mode = VERBOSE;
  struct timeval tv;
  struct timezone tz;
  string mailfilterrc, logfile;
  time_t now;
  string prefsFile;
  extern char *optarg;

  static struct option long_options[] = {
    {"help", 0, NULL, 'h'},
    {"verbose", 0, NULL, 'v'},
    {"silent", 0, NULL, 's'},
    {"mailfilterrc", 1, NULL, 'M'},
    {"logfile", 1, NULL, 'L'},
    {"version", 0, NULL, 'V'},
    {0, 0, 0, 0}
  };

  // Scan command line parameters and options
  while ( (c = getopt_long(argc, argv, "hL:M:sVv", long_options, &option_index)) != -1 ) {
    switch (c) {
    case 'h':
      cout << "Usage: " << PACKAGE << " [OPTION] ..." << endl;
      cout << "Filter your e-mail, get rid of spam." << endl;
      cout << endl;
      cout << "Options:" << endl;
      cout << "  -h, --help                    Display help information and exit" << endl;
      cout << "  -L FILE, --logfile=FILE       Specify logfile location" << endl;
      cout << "  -M FILE, --mailfilterrc=FILE  Specify .mailfilterrc file to use" << endl;
      cout << "  -s, --silent                  Suppress any warnings and messages" << endl;
      cout << "  -v, --verbose                 Increase the amount of output given" << endl;
      cout << "  -V, --version                 Display version information and exit." << endl;
      cout << endl;
      cout << "Report bugs to <baueran@users.sourceforge.net>." << endl;
      exit(0);
      break;
    case 'v':
      mode = VERBOSE;
      break;
    case 's':
      mode = SILENT;
      break;
    case 'V':
      cout << PACKAGE << " " << VERSION << endl;
      cout << endl;
      cout << "Copyright (c) 2000  Andreas Bauer." << endl;
      cout << endl;
      cout << "This is free software; see the source for copying conditions.  There is NO" << endl;
      cout << "warranty; not even for MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE." << endl;
      exit(0);
      break;
    case 'M':
      mailfilterrc = optarg;
      break;
    case 'L':
      logfile = optarg;
      break;
    default:
      // Unrecognized option
      cerr << "Try '" << argv[0] << " --help' for more information." << endl;
      exit(-1);
    }
  }

  // Locate .mailfilterrc depending on whether the user gave one on the command line or not
  if (!mailfilterrc.length()) {
    if (getenv("HOME"))
      prefsFile = (string)getenv("HOME") + (string)"/.mailfilterrc";
    else
      prefsFile = (string)"/home/" + (string)getpwuid(getuid())->pw_name + (string)"/.mailfilterrc";
  }
  else
    prefsFile = mailfilterrc.c_str();

  try {
    // Set some preferences
    pref::Preferences prefs(prefsFile);
    prefs.setMode(mode);
    if (logfile.length())
      prefs.setLogfile(logfile.c_str());
    
    // Prepare the mailfilter Feedback module for error messages and general output
    fb::Feedback report(prefs.getLogfile(), prefs.getMode());
    vector<pref::serverInfo>::iterator curServer = prefs.getServers()->begin();

    // Create separate accounts that can also be checked separately
    while(curServer != prefs.getServers()->end()) {
      pop::PopAccount curAccount(curServer->name, curServer->user, curServer->pass, curServer->port, &prefs);
      gettimeofday(&tv, &tz);
      now = tv.tv_sec;
      report.message((string)PACKAGE + (string)": " + (string)VERSION + (string)" querying " +
		     (string)curServer->name.c_str() + (string)" on " + (string)ctime(&now));
      
      // Now log into account and kill all the spam
      error = curAccount.check();

      if (error == -1)
	report.message((string)PACKAGE + (string)": Error: Could not login to server " + (string)curServer->name.c_str() + (string)". Unknown error.\n");
      else if (error == DNS_LOOKUP_FAILURE)
	report.message((string)PACKAGE + (string)": Error: DNS lookup failure for " + (string)curServer->name.c_str() + (string)". Login failed.\n");
      else if (error == SOCKET_CONNECTION_FAILURE)
	report.message((string)PACKAGE + (string)": Error: Could not create socket. Login failed.\n");
      else if (error == AUTHENTICATION_FAILURE)
	report.message((string)PACKAGE + (string)": Error: Authentication failed. Login canceled.\n");
      else if (error == SOCKET_COMMUNICATION_FAILURE)
	report.message((string)PACKAGE + (string)": Error: Could not establish server connection.\n");
      else if (error == CMD_STAT_FAILED)
	report.message((string)PACKAGE + (string)": Error: Sent STAT, but server responded with an error.\n");
      else if (error == CMD_TOP_FAILED)
	report.message((string)PACKAGE + (string)": Error: Sent TOP, but server responded with an error.\n");
      else if (error == CMD_LIST_FAILED)
	report.message((string)PACKAGE + (string)": Error: Sent LIST, but server responded with an error.\n");
      else if (error == CMD_DELE_FAILED)
	report.message((string)PACKAGE + (string)": Error: Sent DELE, but server responded with an error. Delete failed.\n");
      else if (error == MALFORMED_HEADER_FAILURE)
	report.message((string)PACKAGE + (string)": Error: Encountered a malformed e-mail header which could not be handled by Mailfilter. Program aborted.\n");
      
      curServer++;
      error = 0;
    }
  }
  catch (pref::MalformedPrefsFile) {
    cerr << PACKAGE << ": Error: Preferences file has wrong number of arguments or malformed syntax." << endl;
    return(-1);
  }
  catch (pref::IOException) {
    cerr << PACKAGE << ": Error: Preferences file (usually /home/<username>/.mailfilterrc) could not be read." << endl;
    return(-1);
  }
  catch (fb::LogfileAccess) {
    cerr << PACKAGE << ": Error: Could not access the logfile. Check the permissions of your logfile and look in your .mailfilterrc for the correct file path." << endl;
    return(-1);
  }
  catch (pop::IOException) {
    cerr << PACKAGE << ": Error: Communication failure. Either the server closed your connection due to a time-out problem, or the entire network connection has dropped." << endl;
    return(-1);
  }
  catch (pop::UnknownError) {
    cerr << PACKAGE << ": Error: Unknown communication error occured. Please consider reporting a bug. Thank you." << endl;
    return(-1);    
  }
  catch (...) {
    cerr << PACKAGE << ": Error: This is an unknown error. Please consider reporting a bug. Thank you." << endl;
    return(-1);    
  }

  return 0;
}
